# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Module containing multi-controlled circuits synthesis with and without ancillary qubits."""

from __future__ import annotations
from math import ceil
import numpy as np

from qiskit.exceptions import QiskitError
from qiskit.circuit.quantumcircuit import QuantumCircuit, QuantumRegister, AncillaRegister
from qiskit.circuit.library.standard_gates import (
    HGate,
    CU1Gate,
)

from qiskit._accelerate.synthesis.multi_controlled import (
    c3x as c3x_rs,
    c4x as c4x_rs,
    synth_mcx_n_dirty_i15 as synth_mcx_n_dirty_i15_rs,
    synth_mcx_noaux_v24 as synth_mcx_noaux_v24_rs,
)


def synth_mcx_n_dirty_i15(
    num_ctrl_qubits: int,
    relative_phase: bool = False,
    action_only: bool = False,
) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls based on the paper
    by Iten et al. [1].

    For :math:`k\ge 4` the method uses :math:`k - 2` dirty ancillary qubits, producing a circuit
    with :math:`2 * k - 1` qubits and at most :math:`8 * k - 6` CX gates. For :math:`k\le 3`
    explicit efficient circuits are used instead.

    Args:
        num_ctrl_qubits: The number of control qubits.

        relative_phase: when set to ``True``, the method applies the optimized multi-controlled X gate
            up to a relative phase, in a way that, by lemma 8 of [1], the relative
            phases of the ``action part`` cancel out with the phases of the ``reset part``.

        action_only: when set to ``True``, the method applies only the ``action part`` of lemma 8 of [1].

    Returns:
        The synthesized quantum circuit.

    References:
        1. Iten et. al., *Quantum Circuits for Isometries*, Phys. Rev. A 93, 032318 (2016),
           `arXiv:1501.06911 <http://arxiv.org/abs/1501.06911>`_
    """
    return QuantumCircuit._from_circuit_data(
        synth_mcx_n_dirty_i15_rs(num_ctrl_qubits, relative_phase, action_only)
    )


def synth_mcx_n_clean_m15(num_ctrl_qubits: int) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls using :math:`k - 2`
    clean ancillary qubits with producing a circuit with :math:`2 * k - 1` qubits
    and at most :math:`6 * k - 6` CX gates, by Maslov [1].

    Args:
        num_ctrl_qubits: The number of control qubits.

    Returns:
        The synthesized quantum circuit.

    References:
        1. Maslov., Phys. Rev. A 93, 022311 (2016),
           `arXiv:1508.03273 <https://arxiv.org/pdf/1508.03273>`_
    """

    num_qubits = 2 * num_ctrl_qubits - 1
    q = QuantumRegister(num_qubits, name="q")
    qc = QuantumCircuit(q, name="mcx_vchain")
    q_controls = q[:num_ctrl_qubits]
    q_target = q[num_ctrl_qubits]
    q_ancillas = q[num_ctrl_qubits + 1 :]

    qc.rccx(q_controls[0], q_controls[1], q_ancillas[0])
    i = 0
    for j in range(2, num_ctrl_qubits - 1):
        qc.rccx(q_controls[j], q_ancillas[i], q_ancillas[i + 1])

        i += 1

    qc.ccx(q_controls[-1], q_ancillas[i], q_target)

    for j in reversed(range(2, num_ctrl_qubits - 1)):
        qc.rccx(q_controls[j], q_ancillas[i - 1], q_ancillas[i])

        i -= 1

    qc.rccx(q_controls[0], q_controls[1], q_ancillas[i])

    return qc


def synth_mcx_1_clean_b95(num_ctrl_qubits: int) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls using a single
    clean ancillary qubit producing a circuit with :math:`k + 2` qubits and at most
    :math:`16 * k - 24` CX gates, by [1], [2].

    Args:
        num_ctrl_qubits: The number of control qubits.

    Returns:
        The synthesized quantum circuit.

    References:
        1. Barenco et. al., *Elementary gates for quantum computation*, Phys.Rev. A52 3457 (1995),
           `arXiv:quant-ph/9503016 <https://arxiv.org/abs/quant-ph/9503016>`_
        2. Iten et. al., *Quantum Circuits for Isometries*, Phys. Rev. A 93, 032318 (2016),
           `arXiv:1501.06911 <http://arxiv.org/abs/1501.06911>`_
    """

    if num_ctrl_qubits == 3:
        return synth_c3x()

    elif num_ctrl_qubits == 4:
        return synth_c4x()

    num_qubits = num_ctrl_qubits + 2
    q = QuantumRegister(num_qubits, name="q")
    qc = QuantumCircuit(q, name="mcx_recursive")

    num_ctrl_qubits = len(q) - 1
    q_ancilla = q[-1]
    q_target = q[-2]
    middle = ceil(num_ctrl_qubits / 2)

    # The contruction involving 4 MCX gates is described in Lemma 7.3 of [1], and also
    # appears as Lemma 9 in [2]. The optimization that the first and third MCX gates
    # can be synthesized up to relative phase follows from Lemma 7 in [2], as a diagonal
    # gate following the first MCX gate commutes with the second MCX gate, and
    # thus cancels with the inverse diagonal gate preceding the third MCX gate. The
    # same optimization cannot be applied to the second MCX gate, since a diagonal
    # gate following the second MCX gate would not satisfy the preconditions of Lemma 7,
    # and would not necessarily commute with the third MCX gate.
    controls1 = [*q[:middle]]
    mcx1 = synth_mcx_n_dirty_i15(num_ctrl_qubits=len(controls1), relative_phase=True)
    qubits1 = [*controls1, q_ancilla, *q[middle : middle + mcx1.num_qubits - len(controls1) - 1]]

    controls2 = [*q[middle : num_ctrl_qubits - 1], q_ancilla]
    mcx2 = synth_mcx_n_dirty_i15(num_ctrl_qubits=len(controls2))
    qc2_qubits = [*controls2, q_target, *q[0 : mcx2.num_qubits - len(controls2) - 1]]

    qc.compose(mcx1, qubits1, inplace=True)
    qc.compose(mcx2, qc2_qubits, inplace=True)
    qc.compose(mcx1.inverse(), qubits1, inplace=True)
    qc.compose(mcx2, qc2_qubits, inplace=True)

    return qc


def synth_mcx_gray_code(num_ctrl_qubits: int) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls using the Gray code.

    Produces a quantum circuit with :math:`k + 1` qubits. This method
    produces exponentially many CX gates and should be used only for small
    values of :math:`k`.

    Args:
        num_ctrl_qubits: The number of control qubits.

    Returns:
        The synthesized quantum circuit.
    """
    from qiskit.circuit.library.standard_gates.u3 import _gray_code_chain

    num_qubits = num_ctrl_qubits + 1
    q = QuantumRegister(num_qubits, name="q")
    qc = QuantumCircuit(q, name="mcx_gray")
    qc._append(HGate(), [q[-1]], [])
    scaled_lam = np.pi / (2 ** (num_ctrl_qubits - 1))
    bottom_gate = CU1Gate(scaled_lam)
    definition = _gray_code_chain(q, num_ctrl_qubits, bottom_gate)
    for instr, qargs, cargs in definition:
        qc._append(instr, qargs, cargs)
    qc._append(HGate(), [q[-1]], [])
    return qc


def synth_mcx_noaux_v24(num_ctrl_qubits: int) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls based on
    the implementation for MCPhaseGate.

    In turn, the MCPhase gate uses the decomposition for multi-controlled
    special unitaries described in [1].

    Produces a quantum circuit with :math:`k + 1` qubits.
    The number of CX-gates is quadratic in :math:`k`.

    Args:
        num_ctrl_qubits: The number of control qubits.

    Returns:
        The synthesized quantum circuit.

    References:
        1. Vale et. al., *Circuit Decomposition of Multicontrolled Special Unitary
           Single-Qubit Gates*, IEEE TCAD 43(3) (2024),
           `arXiv:2302.06377 <https://arxiv.org/abs/2302.06377>`_
    """
    circ = QuantumCircuit._from_circuit_data(synth_mcx_noaux_v24_rs(num_ctrl_qubits))
    return circ


def _n_parallel_ccx_x(n: int, apply_x: bool = True) -> QuantumCircuit:
    r"""
    Construct a quantum circuit for creating n-condionally clean ancillae using 3n qubits. This
    implements Fig. 4a of [1]. The circuit applies n relative CCX (RCCX) gates . If apply_x is True,
    each RCCX gate is preceded by an X gate on the target qubit. The order of returned qubits is
    qr_a, qr_b, qr_target.

    Args:
        n: Number of conditionally clean ancillae to create.
        apply_x: If True, apply X gate to the target qubit.

    Returns:
        QuantumCircuit: The quantum circuit for creating n-conditionally clean ancillae.

    References:
        1. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    n_qubits = 3 * n
    q = QuantumRegister(n_qubits, name="q")
    qc = QuantumCircuit(q, name=f"ccxn_{n}")
    qr_a, qr_b, qr_target = q[:n], q[n : 2 * n], q[2 * n :]

    if apply_x:
        qc.x(qr_target)

    qc.rccx(qr_a, qr_b, qr_target)

    return qc


def _linear_depth_ladder_ops(num_ladder_qubits: int) -> tuple[QuantumCircuit, list[int]]:
    r"""
    Helper function to create linear-depth ladder operations used in Khattar and Gidney's MCX synthesis.
    In particular, this implements Step-1 and Step-2 on Fig. 3 of [1] except for the first and last
    CCX gates.

    Args:
        num_ladder_qubits: No. of qubits involved in the ladder operation.

    Returns:
        A tuple consisting of the linear-depth ladder circuit and the index of control qubit to
        apply the final CCX gate.

    Raises:
        QiskitError: If num_ladder_qubits <= 2.

    References:
        1. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    if num_ladder_qubits <= 2:
        raise QiskitError("n_ctrls >= 3 to use MCX ladder. Otherwise, use CCX")

    n = num_ladder_qubits + 1
    qc = QuantumCircuit(n)
    qreg = list(range(n))

    # up-ladder
    for i in range(2, n - 2, 2):
        qc.rccx(qreg[i + 1], qreg[i + 2], qreg[i])
        qc.x(qreg[i])

    # down-ladder
    if n % 2 != 0:
        a, b, target = n - 3, n - 5, n - 6
    else:
        a, b, target = n - 1, n - 4, n - 5

    if target > 0:
        qc.rccx(qreg[a], qreg[b], qreg[target])
        qc.x(qreg[target])

    for i in range(target, 2, -2):
        qc.rccx(qreg[i], qreg[i - 1], qreg[i - 2])
        qc.x(qreg[i - 2])

    mid_second_ctrl = 1 + max(0, 6 - n)
    final_ctrl = qreg[mid_second_ctrl] - 1
    return qc, final_ctrl


def synth_mcx_1_kg24(num_ctrl_qubits: int, clean: bool = True) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls using :math:`1` ancillary qubit as
    described in Sec. 5 of [1].

    Args:
        num_ctrl_qubits: The number of control qubits.
        clean: If True, the ancilla is clean, otherwise it is dirty.

    Returns:
        The synthesized quantum circuit.

    Raises:
        QiskitError: If num_ctrl_qubits <= 2.

    References:
        1. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    if num_ctrl_qubits <= 2:
        raise QiskitError("kg24 synthesis requires at least 3 control qubits. Use CCX directly.")

    q_controls = QuantumRegister(num_ctrl_qubits, name="ctrl")
    q_target = QuantumRegister(1, name="targ")
    q_ancilla = AncillaRegister(1, name="anc")
    qc = QuantumCircuit(q_controls, q_target, q_ancilla, name="mcx_linear_depth")

    ladder_ops, final_ctrl = _linear_depth_ladder_ops(num_ctrl_qubits)

    qc.rccx(q_controls[0], q_controls[1], q_ancilla[0])  #              # create cond. clean ancilla
    qc.compose(ladder_ops, q_ancilla[:] + q_controls[:], inplace=True)  # up-ladder
    qc.ccx(q_ancilla, q_controls[final_ctrl], q_target)  #              # target
    qc.compose(  #                                                      # down-ladder
        ladder_ops.inverse(),
        q_ancilla[:] + q_controls[:],
        inplace=True,
    )
    qc.rccx(q_controls[0], q_controls[1], q_ancilla[0])  #              # undo cond. clean ancilla

    if not clean:
        # perform toggle-detection if ancilla is dirty
        qc.compose(ladder_ops, q_ancilla[:] + q_controls[:], inplace=True)
        qc.ccx(q_ancilla, q_controls[final_ctrl], q_target)
        qc.compose(ladder_ops.inverse(), q_ancilla[:] + q_controls[:], inplace=True)

    return qc


def synth_mcx_1_clean_kg24(num_ctrl_qubits: int) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls using :math:`1` clean ancillary qubit
    producing a circuit with :math:`2k-3` Toffoli gates and depth :math:`O(k)` as described in
    Sec. 5.1 of [1].

    Args:
        num_ctrl_qubits: The number of control qubits.

    Returns:
        The synthesized quantum circuit.

    Raises:
        QiskitError: If num_ctrl_qubits <= 2.

    References:
        1. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    return synth_mcx_1_kg24(num_ctrl_qubits, clean=True)


def synth_mcx_1_dirty_kg24(num_ctrl_qubits: int) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls using :math:`1` dirty ancillary qubit
    producing a circuit with :math:`4k-8` Toffoli gates and depth :math:`O(k)` as described in
    Sec. 5.3 of [1].

    Args:
        num_ctrl_qubits: The number of control qubits.

    Returns:
        The synthesized quantum circuit.

    Raises:
        QiskitError: If num_ctrl_qubits <= 2.

    References:
        1. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    return synth_mcx_1_kg24(num_ctrl_qubits, clean=False)


def _build_logn_depth_ccx_ladder(
    ancilla_idx: int, ctrls: list[int], skip_cond_clean: bool = False
) -> tuple[QuantumCircuit, list[int]]:
    r"""
    Helper function to build a log-depth ladder compose of CCX and X gates as shown in Fig. 4b of [1].

    Args:
        ancilla_idx: Index of the ancillary qubit.
        ctrls: List of control qubits.
        skip_cond_clean: If True, do not include the conditionally clean ancilla (step 1 and 5 in
            Fig. 4b of [1]).

    Returns:
        A tuple consisting of the log-depth ladder circuit of conditionally clean ancillae and the
        list of indices of control qubit to apply the linear-depth MCX gate.

    Raises:
        QiskitError: If no. of qubits in parallel CCX + X gates are not the same.

    References:
        1. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    qc = QuantumCircuit(len(ctrls) + 1)
    anc = [ancilla_idx]
    final_ctrls = []

    while len(ctrls) > 1:
        next_batch_len = min(len(anc) + 1, len(ctrls))
        ctrls, nxt_batch = ctrls[next_batch_len:], ctrls[:next_batch_len]
        new_anc = []
        while len(nxt_batch) > 1:
            ccx_n = len(nxt_batch) // 2
            st = int(len(nxt_batch) % 2)
            ccx_x, ccx_y, ccx_t = (
                nxt_batch[st : st + ccx_n],
                nxt_batch[st + ccx_n :],
                anc[-ccx_n:],
            )
            if not len(ccx_x) == len(ccx_y) == ccx_n >= 1:
                raise QiskitError(
                    f"Invalid CCX gate parameters: {len(ccx_x)=} != {len(ccx_y)=} != {len(ccx_n)=}"
                )
            if ccx_t != [ancilla_idx]:
                qc.compose(_n_parallel_ccx_x(ccx_n), ccx_x + ccx_y + ccx_t, inplace=True)
            else:
                if not skip_cond_clean:
                    qc.rccx(ccx_x[0], ccx_y[0], ccx_t[0])  # # create conditionally clean ancilla

            new_anc += nxt_batch[st:]  #                     # newly created cond. clean ancilla
            nxt_batch = ccx_t + nxt_batch[:st]
            anc = anc[:-ccx_n]

        anc = sorted(anc + new_anc)
        final_ctrls += nxt_batch

    final_ctrls += ctrls
    final_ctrls = sorted(final_ctrls)
    return qc, final_ctrls[:-1]  # exclude ancilla


def synth_mcx_2_kg24(num_ctrl_qubits: int, clean: bool = True) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls using :math:`2` ancillary qubits.
    as described in Sec. 5 of [1].

    Args:
        num_ctrl_qubits: The number of control qubits.
        clean: If True, the ancilla is clean, otherwise it is dirty.

    Returns:
        The synthesized quantum circuit.

    Raises:
        QiskitError: If num_ctrl_qubits <= 2.

    References:
        1. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    if num_ctrl_qubits <= 2:
        raise QiskitError("kg24 synthesis requires at least 3 control qubits. Use CCX directly.")

    q_control = QuantumRegister(num_ctrl_qubits, name="ctrl")
    q_target = QuantumRegister(1, name="targ")
    q_ancilla = AncillaRegister(2, name="anc")
    qc = QuantumCircuit(q_control, q_target, q_ancilla, name="mcx_logn_depth")

    ladder_ops, final_ctrls = _build_logn_depth_ccx_ladder(
        num_ctrl_qubits, list(range(num_ctrl_qubits))
    )
    qc.compose(ladder_ops, q_control[:] + [q_ancilla[0]], inplace=True)
    if len(final_ctrls) == 1:  # Already a toffoli
        qc.ccx(q_ancilla[0], q_control[final_ctrls[0]], q_target)
    else:
        mid_mcx = synth_mcx_1_clean_kg24(len(final_ctrls) + 1)
        qc.compose(
            mid_mcx,
            [q_ancilla[0]]
            + q_control[final_ctrls]
            + q_target[:]
            + [q_ancilla[1]],  # ctrls, targ, anc
            inplace=True,
        )
    qc.compose(ladder_ops.inverse(), q_control[:] + [q_ancilla[0]], inplace=True)

    if not clean:
        # perform toggle-detection if ancilla is dirty
        ladder_ops_new, final_ctrls = _build_logn_depth_ccx_ladder(
            num_ctrl_qubits, list(range(num_ctrl_qubits)), skip_cond_clean=True
        )
        qc.compose(ladder_ops_new, q_control[:] + [q_ancilla[0]], inplace=True)
        if len(final_ctrls) == 1:
            qc.ccx(q_ancilla[0], q_control[final_ctrls[0]], q_target)
        else:
            qc.compose(
                mid_mcx,
                [q_ancilla[0]] + q_control[final_ctrls] + q_target[:] + [q_ancilla[1]],
                inplace=True,
            )
        qc.compose(ladder_ops_new.inverse(), q_control[:] + [q_ancilla[0]], inplace=True)

    return qc


def synth_mcx_2_clean_kg24(num_ctrl_qubits: int) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls using :math:`2` clean ancillary qubits
    producing a circuit with :math:`2k-3` Toffoli gates and depth :math:`O(\log(k))` as described in
    Sec. 5.2 of [1].

    Args:
        num_ctrl_qubits: The number of control qubits.

    Returns:
        The synthesized quantum circuit.

    Raises:
        QiskitError: If num_ctrl_qubits <= 2.

    References:
        1. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    return synth_mcx_2_kg24(num_ctrl_qubits, clean=True)


def synth_mcx_2_dirty_kg24(num_ctrl_qubits: int) -> QuantumCircuit:
    r"""
    Synthesize a multi-controlled X gate with :math:`k` controls using :math:`2` dirty ancillary qubits
    producing a circuit with :math:`4k-8` Toffoli gates and depth :math:`O(\log(k))` as described in
    Sec. 5.4 of [1].

    Args:
        num_ctrl_qubits: The number of control qubits.

    Returns:
        The synthesized quantum circuit.

    Raises:
        QiskitError: If num_ctrl_qubits <= 2.

    References:
        1. Khattar and Gidney, Rise of conditionally clean ancillae for optimizing quantum circuits
        `arXiv:2407.17966 <https://arxiv.org/abs/2407.17966>`__
    """

    return synth_mcx_2_kg24(num_ctrl_qubits, clean=False)


def synth_c3x() -> QuantumCircuit:
    """Efficient synthesis of 3-controlled X-gate."""
    return QuantumCircuit._from_circuit_data(c3x_rs())


def synth_c4x() -> QuantumCircuit:
    """Efficient synthesis of 4-controlled X-gate."""
    return QuantumCircuit._from_circuit_data(c4x_rs())



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Context-aware dynamical decoupling."""

from __future__ import annotations
from enum import Enum

import itertools
import logging

from pprint import pformat
from dataclasses import dataclass, field
from collections import defaultdict
import numpy as np
import rustworkx as rx

from qiskit.circuit import QuantumCircuit, Qubit, Reset, Gate
from qiskit.circuit.library import CXGate, ECRGate, XGate
from qiskit.dagcircuit import DAGCircuit
from qiskit.converters import circuit_to_dag
from qiskit.dagcircuit import DAGOpNode, DAGInNode
from qiskit.transpiler.target import Target
from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.coupling import CouplingMap
from qiskit.transpiler.exceptions import TranspilerError

from .pad_delay import PadDelay

logger = logging.getLogger(__name__)


class ContextAwareDynamicalDecoupling(TransformationPass):
    """Implement an X-sequence dynamical decoupling considering the gate- and qubit-context.

    This pass implements a context-aware dynamical decoupling (DD) [1], which ensures that

        (1) simultaneously occurring DD sequences on device-adjacent qubits are mutually orthogonal, and
        (2) DD sequences on spectator qubits of ECR and CX gates are orthogonal to the echo
            pulses on the neighboring target/control qubits.

    The mutually orthogonal DD sequences are currently Walsh-Hadamard sequences, consisting of only
    X gates. In some cases it might therefore be beneficial to use :class:`.PadDynamicalDecoupling`
    with more generic sequences, such as XY4.

    This pass performs best if the two-qubit interactions have the same durations on the
    device, as it allows to align delay sequences and take into account potential control and target
    operations on neighboring qubits. However, it is still valid if this is not the case.

    .. note::

        If this pass is run within a pass manager (as in the example below), it will
        automatically run :class:`.PadDelay` to allocate the delays. If instead it is run as
        standalone (not inside a :class:`.PassManager`), the delays must already be inserted.


    Example::

        from qiskit.circuit import QuantumCircuit
        from qiskit.circuit.library import QFTGate
        from qiskit.transpiler import PassManager
        from qiskit.transpiler.passes import ALAPScheduleAnalysis, ContextAwareDynamicalDecoupling
        from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
        from qiskit_ibm_runtime.fake_provider import FakeSherbrooke

        num_qubits = 10
        circuit = QuantumCircuit(num_qubits)
        circuit.append(QFTGate(num_qubits), circuit.qubits)
        circuit.measure_all()

        target = FakeSherbrooke().target

        pm = generate_preset_pass_manager(optimization_level=2, target=target)
        dd = PassManager([
            ALAPScheduleAnalysis(target=target),
            ContextAwareDynamicalDecoupling(target=target),
        ])

        transpiled = pm.run(circuit)
        with_dd = dd.run(transpiled)

        print(with_dd.draw(idle_wires=False))

    References:

        [1] A. Seif et al. (2024). Suppressing Correlated Noise in Quantum Computers via
            Context-Aware Compiling, `arXiv:2403.06852 <https://arxiv.org/abs/2403.06852>`_.

    """

    def __init__(
        self,
        target: Target,
        *,
        min_duration: int | None = None,
        skip_reset_qubits: bool = True,
        skip_dd_threshold: float = 1.0,
        pulse_alignment: int | None = None,
        coloring_strategy: rx.ColoringStrategy = rx.ColoringStrategy.Saturation,
    ) -> None:
        """
        Args:
            target: The :class:`.Target` of the device to run the circuit.
            min_duration: Minimal delay duration (in ``dt``) to insert a DD sequence. This
                can be useful, e.g. if a big delay block would be interrupted and split into
                smaller blocks due to a very short, adjacent delay. If ``None``, this is set
                to be at least twice the difference of the longest/shortest CX or ECR gate.
            skip_reset_qubits: Skip initial delays and delays after a reset.
            skip_dd_threshold: Skip dynamical decoupling on an idle qubit, if the duration of
                the decoupling sequence exceeds this fraction of the idle window. For example, to
                skip a DD sequence if it would take up more than 95% of the idle time, set this
                value to 0.95. A value of 1. means that the DD sequence is applied if it fits into
                the window.
            pulse_alignment: The hardware constraints (in ``dt``) for gate timing allocation.
                If provided, the inserted gates will only be executed on integer multiples of
                this value. This is usually provided on ``backend.configuration().timing_constraints``.
                If ``None``, this is extracted from the ``target``.
            coloring_strategy: The coloring strategy used for ``rx.greedy_graph_color``.
                Defaults to a saturation strategy, which is optimal on bipartite graphs,
                see Section 1.2.2.8 of [2].

        References:

            [2] A. Kosowski, and K. Manuszewski, Classical Coloring of Graphs, Graph Colorings,
                2-19, 2004. ISBN 0-8218-3458-4.

        """
        super().__init__()

        if not 0 <= skip_dd_threshold <= 1:
            raise ValueError(f"skip_dd_threshold must be in [0, 1], but is {skip_dd_threshold}")

        if min_duration is None:
            if target.dt is None:
                min_duration = 0
            else:
                min_duration = 2 * _gate_length_variance(target) / target.dt

        self._min_duration = min_duration
        self._skip_reset_qubits = skip_reset_qubits
        self._skip_dd_threshold = skip_dd_threshold
        self._target = target
        self._coupling_map = target.build_coupling_map()  # build once and re-use for performance
        self._pulse_alignment = (
            target.pulse_alignment if pulse_alignment is None else pulse_alignment
        )
        self._coloring_strategy = coloring_strategy
        self._sequence_generator = WalshHadamardSequence()

        # Use PadDelay to insert Delay operations into the DAG before running this pass.
        # This could be integrated into this pass as well, saving one iteration over the DAG,
        # but this would repeat logic and is currently no bottleneck.
        self.requires = [PadDelay(target=target)]

    def run(self, dag: DAGCircuit) -> DAGCircuit:
        # check the schedule analysis was run
        if "node_start_time" not in self.property_set:
            raise RuntimeError(
                "node_start_time not found in the pass manager's property set. "
                "Please run a scheduling pass before calling ContextAwareDynamicalDecoupling."
            )

        # 1) Find all delay instructions that qualify. We split them into explicit
        # "begin" and "end" events and sort them.
        sorted_delay_events = self._get_sorted_delays(dag)

        # 2) Identify adjacent delays by overlap in time and adjacency on device.
        qubit_map = {bit: idx for idx, bit in enumerate(dag.qubits)}  # map qubits to their indices
        adjacent_delay_blocks = self._collect_adjacent_delay_blocks(sorted_delay_events, qubit_map)
        logger.debug("adjacent delay blocks: %s", pformat(adjacent_delay_blocks))

        # 3) Split the blocks into multi-qubit delay layers
        merged_delays = []
        for block in adjacent_delay_blocks:
            merged_delays_in_block = self._split_adjacent_block(block, qubit_map)
            merged_delays += merged_delays_in_block

        # 4) For each multi-qubit delay, find a coloring (that takes into account neighboring
        # CX or ECR gates) and get the DD sequence
        all_delays = set()  # keep a list of all delays to iterate over them easily later

        for merged_delay in merged_delays:
            all_delays.update(merged_delay.ops)

            # get coloring inside the n-qubit delay
            coloring = self._get_wire_coloring(dag, merged_delay)

            # get the DD sequence on the qubit
            duration = merged_delay.end - merged_delay.start
            checked_durations_cache = set()
            for op in merged_delay.ops:
                dd_circuit, start_times = self._get_dd_sequence(
                    coloring[op.index],
                    op.index,
                    duration,
                    checked_durations_cache,
                )
                op.replacement.compose(dd_circuit, inplace=True, copy=False)
                if len(op.start_times) == 0:
                    op.start_times += start_times
                else:
                    start_times = [op.start_times[-1] + time for time in start_times]

                op.start_times += start_times

        # 5) Replace each delay operation with its individual DD sequence
        qubit_map = dict(enumerate(dag.qubits))
        for delay in all_delays:
            # replace it with its stored replacement
            as_dag = circuit_to_dag(delay.replacement)
            node_ids = [node._node_id for node in as_dag.topological_op_nodes()]
            id_map = dag.substitute_node_with_dag(
                delay.op, as_dag, {as_dag.qubits[0]: qubit_map[delay.index]}
            )

            # update node start times
            for node_id, start_time in zip(node_ids, delay.start_times):
                self.property_set["node_start_time"][id_map[node_id]] = start_time

        return dag

    def get_orthogonal_sequence(self, order: int) -> tuple[list[float], list[Gate]]:
        """Return a DD sequence of given order, where different orders are orthogonal."""
        spacing = self._sequence_generator.get_sequence(order)
        return spacing, [XGate() for _ in spacing[:-1]]

    def _get_sorted_delays(
        self,
        dag: DAGCircuit,
    ) -> list[DelayEvent]:
        """Get sorted DelayEvent objects for all eligible delay operations."""

        def is_after_reset(node):
            if not self._skip_reset_qubits:
                return False  # if we do not skip reset qubits, we always want to use the delay

            predecessor = next(dag.predecessors(node))  # a delay has one 1 predecessor
            return isinstance(predecessor, DAGInNode) or isinstance(predecessor.op, Reset)

        qubit_map = {bit: index for index, bit in enumerate(dag.qubits)}

        eligible_delays = [
            DelayEvent(
                event_type,
                (
                    start_time
                    if event_type == EventType.BEGIN
                    else start_time + self._duration(node, qubit_map)
                ),
                node,
            )
            for node, start_time in self.property_set["node_start_time"].items()
            if (
                node.op.name == "delay"
                and not is_after_reset(node)
                and self._duration(node, qubit_map) > self._min_duration
            )
            for event_type in (EventType.BEGIN, EventType.END)
        ]

        sorted_events = sorted(eligible_delays, key=DelayEvent.sort_key)
        logger.debug("Sorted delay events: %s", sorted_events)

        return sorted_events

    def _get_wire_coloring(self, dag: DAGCircuit, merged_delay: MultiDelayOp) -> dict[int, int]:
        """Find a wire coloring for a multi-delay operation.

        This function returns a dictionary that includes the coloring (as int) for the indices in the
        ``merged_delay`` as ``{index: color}`` pairs. Spectator qubits are handled by assigning
        neighboring qubits with a CX or ECR a color (0 if control, 1 if target) and including them
        in the coloring problem.
        """
        # get neighboring wires, for which we will give initial colors
        neighbors = set()
        index_map = dict(enumerate(dag.qubits))
        qubit_map = {bit: index for index, bit in enumerate(dag.qubits)}

        for delay_op in merged_delay.ops:
            # use coupling_map.graph.neighbors_undirected once Qiskit/rustworkx#1254 is in a release
            new_neighbors = {
                i
                for i in range(dag.num_qubits())
                if self._coupling_map.distance(i, delay_op.index) == 1
            }
            # do not add indices that are already in the merged delay
            neighbors.update(new_neighbors.difference(merged_delay.indices))

        # build a subgraph we will apply the coloring function on
        wires = sorted(neighbors.union(merged_delay.indices))
        subgraph = self._coupling_map.graph.subgraph(list(wires)).to_undirected()
        glob2loc = dict(zip(wires, subgraph.node_indices()))
        preset_coloring = {index: None for index in subgraph.node_indices()}

        # find the neighbor wires and check if ctrl/tgt spectator
        for wire in neighbors:
            for op_node in dag.nodes_on_wire(dag.qubits[wire], only_ops=True):
                # check if the operation occurs during the delay
                op_start = self.property_set["node_start_time"][op_node]
                op_end = op_start + self._duration(op_node, qubit_map)
                if (
                    isinstance(op_node.op, (CXGate, ECRGate))
                    and op_start < merged_delay.end
                    and op_end > merged_delay.start
                ):
                    # set coloring to 0 if ctrl, and to 1 if tgt
                    ctrl, tgt = op_node.qargs
                    if index_map[wire] == ctrl:
                        preset_coloring[glob2loc[wire]] = 0
                    if index_map[wire] == tgt:
                        preset_coloring[glob2loc[wire]] = 1

        local_coloring = rx.graph_greedy_color(
            subgraph,
            preset_color_fn=lambda loc: preset_coloring[loc],
            strategy=self._coloring_strategy,
        )

        # map the local indices of the subgraph back to our DAG indices
        loc2glob = dict(zip(subgraph.node_indices(), wires))
        coloring = {loc2glob[loc]: color for loc, color in local_coloring.items()}

        # for debugging purposes, print the coloring of each block
        logger.debug("Coloring for block %s: %s", merged_delay, coloring)

        return coloring

    def _get_dd_sequence(
        self,
        order: int,
        index: int,
        duration: int,
        checked_durations_cache: set[int],
    ) -> tuple[QuantumCircuit, list[int]]:
        """Get a DD sequence of specified order on qubit with a given index.

        Takes the gate durations, the pulse alignment and a set to cache, which gate lengths
        we've already checked to match the pulse alignment. Returns the DD sequence and the
        node start times as list.
        """
        instruction_durations = self._target.durations()
        # check the X gate on the active qubit is compatible with pulse alignment
        if index not in checked_durations_cache:
            x_duration = instruction_durations.get("x", index)
            if x_duration % self._pulse_alignment != 0:
                raise TranspilerError(
                    f"X gate length on qubit with index {index} is {x_duration} which is not "
                    f"an integer multiple of the pulse alignment {self._pulse_alignment}."
                )
            checked_durations_cache.add(index)

        spacing, dd_sequence = self.get_orthogonal_sequence(order=order)

        # check if DD can be applied or if there is not enough time
        dd_sequence_duration = sum(instruction_durations.get(gate, index) for gate in dd_sequence)
        slack = duration - dd_sequence_duration
        slack_fraction = slack / duration
        if 1 - slack_fraction >= self._skip_dd_threshold:  # dd doesn't fit
            seq = QuantumCircuit(1)
            seq.delay(duration, 0)
            return seq, [0]

        # compute actual spacings in between the delays, taking into account
        # the pulse alignment restriction of the hardware
        taus = self._constrain_spacing(spacing, slack)

        # apply the DD gates
        # tau has one more entry than the gate sequence
        start_times = []
        time = 0  # track the node start time
        seq = QuantumCircuit(1)  # if the DD sequence has a global phase, add it here
        for tau, gate in itertools.zip_longest(taus, dd_sequence):
            if tau > 0:
                seq.delay(tau, 0)
                start_times.append(time)
                time += tau
            if gate is not None:
                seq.append(gate, [0])
                start_times.append(time)
                time += instruction_durations.get(gate, index)

        return seq, start_times

    def _constrain_spacing(self, spacing, slack):
        def _constrained_length(values):
            return self._pulse_alignment * np.floor(values / self._pulse_alignment)

        taus = _constrained_length(slack * np.asarray(spacing))
        unused_slack = slack - sum(taus)  # unused, due to rounding to int multiples of dt
        middle_index = int((len(taus) - 1) / 2)  # arbitrary: redistribute to middle
        to_middle = _constrained_length(unused_slack)
        taus[middle_index] += to_middle  # now we add up to original delay duration
        if unused_slack - to_middle:
            taus[-1] += unused_slack - to_middle

        return taus

    def _collect_adjacent_delay_blocks(
        self,
        sorted_delay_events: list[DelayEvent],
        qubit_map: dict[Qubit, int],
    ) -> list[AdjacentDelayBlock]:
        """Collect delay events into adjacent blocks.

        Events in an adjacent block are overlapping in time and adjacent on the device.
        See also the dataclass ``AdjacentDelayBlock`` for more information.

        Args:
            sorted_delay_events: All eligible delay operations, sorted by time and type.
            qubit_map: A map from qubit instance to qubit index.

        Returns:
            A list of adjacent delay blocks.
        """
        open_delay_blocks = []
        closed_delay_blocks = []

        def _open_delay_block(delay_event):
            open_delay_blocks.append(
                AdjacentDelayBlock(
                    events=[delay_event], active_qubits=set(delay_event.op_node.qargs)
                )
            )
            return open_delay_blocks[-1]

        def _update_delay_block(open_delay_block, delay_event):
            """Add another delay event to an existing block to either extend or close it."""
            open_delay_block.events.append(delay_event)

            # at this point we know that delay_event.op_node.qargs is active
            open_delay_block.active_qubits -= set(delay_event.op_node.qargs)

            if not open_delay_block.active_qubits:
                open_delay_blocks.remove(open_delay_block)
                closed_delay_blocks.append(open_delay_block)

        def _combine_delay_blocks(delay_blocks):
            survivor, *doomed = delay_blocks

            for doomed_delay_group in doomed:
                # Add events and qubits from doomed block to survivor.
                if logger.isEnabledFor(logging.DEBUG):
                    if survivor.active_qubits.intersection(doomed_delay_group.active_qubits):
                        logger.debug("More than one open delay on a qubit?")

                survivor.events.extend(doomed_delay_group.events)
                survivor.active_qubits.update(doomed_delay_group.active_qubits)

                open_delay_blocks.remove(doomed_delay_group)
                survivor.events.sort(key=DelayEvent.sort_key)  # Maintain sorted event order

        for delay_event in sorted_delay_events:
            # This could be avoided by keeping a map of device qubit to open
            # block and only considering neighbors of current event.
            # use coupling_map.graph.neighbors_undirected once Qiskit/rustworkx#1254 is in a release
            adjacent_open_delay_blocks = [
                open_delay
                for open_delay in open_delay_blocks
                if any(
                    self._coupling_map.distance(
                        qubit_map[delay_event.op_node.qargs[0]], qubit_map[open_delay_qubit]
                    )
                    <= 1
                    for open_delay_qubit in open_delay.active_qubits
                )
            ]

            if delay_event.type == EventType.BEGIN:
                # If crossing a begin edge, check if there are any open delays that are adjacent.
                # If so, add current event to that group.

                if len(adjacent_open_delay_blocks) == 0:
                    # Make a new delay block
                    _open_delay_block(delay_event)
                else:
                    # Make a new block and combine that with adjacent open blocks
                    new_block = _open_delay_block(delay_event)
                    _combine_delay_blocks(adjacent_open_delay_blocks + [new_block])

            else:
                if logger.isEnabledFor(logging.DEBUG):
                    # If crossing a end edge, remove this qubit from the actively delaying qubits"
                    if len(adjacent_open_delay_blocks) != 1:
                        logger.debug("Closing edge w/o an open delay?")

                _update_delay_block(adjacent_open_delay_blocks[0], delay_event)

        # log the open delays and the number of closed delays
        logger.debug("-- post collect")
        logger.debug("open delays: %s", open_delay_blocks)
        logger.debug("len(closed delays): %s", len(closed_delay_blocks))

        # validate the results, there should be no open delays and all active qubits
        # should be accounted for
        if logger.isEnabledFor(logging.DEBUG):
            if len(open_delay_blocks) > 0:
                logger.debug("Failed to close all open delays.")

            for closed_delay in closed_delay_blocks:
                if len(closed_delay.active_qubits) > 0:
                    logger.debug("Failed to remove active qubits on closed delay %s.", closed_delay)

                closed_delay.validate()

        return closed_delay_blocks

    def _split_adjacent_block(
        self,
        adjacent_block: AdjacentDelayBlock,
        qubit_map: dict[Qubit, int],
    ) -> list[MultiDelayOp]:
        """Split adjacent delay blocks in concurrent layers of maximum width.

        This code performs the following steps:

            1) Find the atomic time windows where delays can begin and end.
            2) Find which connected qubit components are jointly idle during the windows.
            3) Merge time-adjacent qubit components.

        """
        # 1) Find times at which a new delay event is happening and we might have to break up
        # the adjacent delays
        breakpoints = []
        # a dictionary of {qubit_index: [delay_op1, delay_op2, ...]}
        all_delays = defaultdict(list)

        for event in adjacent_block.events:
            index = qubit_map[event.op_node.qargs[0]]  # delays are 1q ops
            if event.type == EventType.BEGIN:
                delay = DelayOp(start=event.time, index=index, op=event.op_node)
                all_delays[index].append(delay)
            else:  # find which delay to close
                all_delays[index][-1].end = event.time

            if len(breakpoints) == 0 or event.time > breakpoints[-1]:
                breakpoints.append(event.time)

        # 2) Find which delays are active during which time windows, where a time window
        # is defined as (breakpoints[i], breakpoints[i + 1])
        active_delays = {}  # {window: [group1, group2]} where each group = (index1, index2, ..)
        op_map = {}  # {(qubit_index, window): delay operation as DAGOpNode}
        windows = list(zip(breakpoints[:-1], breakpoints[1:]))
        for window in windows:
            active = []

            # check which delays are active during the window
            # this could be e.g. [0, 1, 2, 5, 6, 9]
            for index, delays in all_delays.items():
                for delay in delays:
                    # since the windows are atomic, we have three cases:
                    # (1) the delay starts at the time window, (2) it ends with it, (3) it contains it
                    if delay.start <= window[0] and delay.end >= window[1]:
                        active.append(index)
                        delay.add_window(window)
                        op_map[(index, window)] = delay

            # check which are adjacent
            # on a linear topology, we would get [[0, 1, 2], [5, 6], [9]]
            visited = defaultdict(lambda: False)
            grouped = []
            for start_index in active:
                if visited[start_index]:
                    continue

                active_neighbors = {start_index}
                _dfs(start_index, self._coupling_map, active_neighbors, active)

                for index in active_neighbors:
                    visited[index] = True

                group = tuple(sorted(active_neighbors))  # must be sorted to merge later
                grouped.append(group)

            # sanity check: groups must be disjoint
            if logger.isEnabledFor(logging.DEBUG):
                for i, g1 in enumerate(grouped):
                    for g2 in grouped[i + 1 :]:
                        if len(set(g1).intersection(g2)) > 0:
                            logger.debug("Groups not disjoint: %s and %s.", g1, g2)

            active_delays[window] = grouped

        # 3) Merge time-adjacent active delays
        merged_delays = []
        open_groups = {}
        for window in windows:
            next_groups = active_delays[window]

            # check which opened groups are still active in this window
            next_open_groups = {}
            for open_group, delay in open_groups.items():
                if open_group in next_groups:
                    # extend the delay operation and remove a breakpoint
                    delay.end = window[1]
                    for op in delay.ops:
                        op.breakpoints.remove(window[0])

                    if logger.isEnabledFor(logging.DEBUG):
                        involved_ops = {op_map[(index, window)] for index in open_group}
                        if len(involved_ops.difference(delay.ops)) > 0:
                            logger.debug("Not all involved operations are part of the joint delay.")

                    next_open_groups[open_group] = delay
                else:
                    # the group has not been extended, close it!
                    merged_delays.append(delay)

            # add the new open groups
            for group in next_groups:
                if group not in open_groups:
                    involved_ops = {op_map[(index, window)] for index in group}
                    next_open_groups[group] = MultiDelayOp(
                        start=window[0], end=window[1], indices=group, ops=involved_ops
                    )

            open_groups = next_open_groups

        # add all from the final layer
        for open_group, delay in open_groups.items():
            merged_delays.append(delay)

        logger.debug("Split adjacent delay block into %s", "\n".join(map(str, merged_delays)))

        return merged_delays

    def _duration(self, node: DAGOpNode, qubit_map: dict[Qubit, int]) -> float:
        # this is cached on the target, so we can repeatedly call it w/o penalty
        instruction_durations = self._target.durations()
        indices = [qubit_map[bit] for bit in node.qargs]

        return instruction_durations.get(node.op, indices)


class EventType(Enum):
    """Delay event type, which is either begin or end."""

    BEGIN = 0
    END = 1


@dataclass
class DelayEvent:
    """Represent a single-qubit delay event, which is either begin or end of a delay instruction."""

    type: EventType
    time: int  # Time in this circuit, in dt
    op_node: DAGOpNode  # The node for the circuit delay

    @staticmethod
    def sort_key(event: DelayEvent) -> tuple(int, int):
        """Sort events, first by time then by type ('end' events come before 'begin' events)."""
        return (
            event.time,  # Sort by event time
            0 if event.type == EventType.END else 1,  # With 'end' events before 'begin'
        )


@dataclass
class DelayOp:
    """Represent a delay operation."""

    start: int
    index: int
    op: DAGOpNode  # the circuit op node this represents
    end: int | None = None  # None means currently unknown
    breakpoints: list[int] = field(
        default_factory=list
    )  # timepoints at which the delay op is split
    # circuit with which we will replace this delay
    start_times: list[int] = field(default_factory=list)
    replacement: QuantumCircuit = field(default_factory=lambda: QuantumCircuit(1))

    def __hash__(self):
        return hash(self.op)

    def add_window(self, window: tuple[int, int]):
        """Add a time window to the delay op.

        This means the delay is active during this window and we add a potential breakpoint.
        """
        if self.end is None:
            raise ValueError("Cannot add a window if DelayOp.end is None. Please set it.")

        start, end = window
        if self.start < start and start not in self.breakpoints:
            self.breakpoints.append(start)
        if self.end > end and end not in self.breakpoints:
            self.breakpoints.append(end)

        self.breakpoints = sorted(self.breakpoints)


@dataclass
class MultiDelayOp:
    """A multi-qubit delay operation."""

    start: int
    end: int
    indices: list[int]
    ops: set[DelayOp]
    replacing: set[DAGOpNode] = field(default_factory=set)

    def __str__(self) -> str:
        return f"MultiDelay({self.start}:{self.end} on {self.indices})"


@dataclass
class AdjacentDelayBlock:
    """Group of circuit delays which are collectively adjacent in time and on device.

    For example, here the 3 delay operations on q0, q1 and q2 form an adjacent delay block.

        q0: -██████---------  |  qubits q0,q1,q2 have adjacent delay
        q1: ------███████---  |  operations, since the delay operations
        q2: --█████████-----  |  all overlap
        q3: -----------████-  -> this delay starts when delay on q2 ends, so they have no overlap
        q4: ----████--------  -> this clearly has no overlap with something else

    """

    events: list[DelayEvent]
    active_qubits: set[Qubit]

    def validate(self, log: bool = True) -> None:
        """Validate the list of delay events in the adjacent block.

        Args:
            log: If ``True`` log invalid blocks on DEBUG level. Otherwise raise an error if the
                block is invalid.

        Raises:
            RuntimeError: If the blocks are not ordered by time and event type.
        """

        def notify(msg, *args):
            if log:
                logger.debug(msg, *args)
            else:
                raise RuntimeError(msg.format(*args))

        for idx, event in enumerate(self.events[:-1]):
            if event.time > self.events[idx + 1].time:
                notify("adjacent_delay_block.events not ordered by time")

            if event.time == self.events[idx + 1].time:
                # At same time, can either be ('begin', 'begin'), ('end', 'begin') or ('end', 'end')
                if (event.type, self.events[idx + 1].type) == (EventType.BEGIN, EventType.END):
                    notify(
                        "Events in the AdjacentDelayBlock are not correctly sorted by "
                        "event type. At same time, we can have either of (begin, begin), "
                        "(end, begin) or (end, end). This happened at time %s.",
                        event.time,
                    )


def _dfs(qubit, cmap: CouplingMap, visited, active_qubits):
    """Depth-first search to get the widest group of idle qubits during a given time frame."""
    # use coupling_map.graph.neighbors_undirected once Qiskit/rustworkx#1254 is in a release
    neighbors = {other for other in active_qubits if cmap.distance(qubit, other) == 1}
    for neighbor in neighbors:
        if neighbor in active_qubits and neighbor not in visited:
            visited.add(neighbor)
            _dfs(neighbor, cmap, visited, active_qubits)


class WalshHadamardSequence:
    """Get Walsh-Hadamard sequences for DD up to arbitrary order."""

    def __init__(self, max_order: int = 5):
        """
        Args:
            max_order: The maximal order for which the sequences are computed.
        """
        # these are set in set_max_order
        self.sequences = None
        self.max_order = None

        self.set_max_order(max_order)

    def set_max_order(self, max_order: int) -> None:
        """Set the maximal available order."""
        if self.max_order is not None:
            if max_order <= self.max_order:
                return

        # get the dimension of the transformation matrix we need,
        # this is given by the smallest power of 2 that includes max_order
        num_krons = int(np.ceil(np.log2(max_order + 1)))
        self.max_order = 2**num_krons - 1

        rows = _get_transformation_matrix(num_krons).tolist()
        distances = [_bitflips_to_timings(row) for row in rows]
        num_flips = [len(distance) - 1 for distance in distances]

        # sort by the number of flips and throw out the first one,
        # which corresponds to the 000... bit-sequence, i.e., no flip
        indices = np.argsort(num_flips)[1:]

        self.sequences = [distances[i] for i in indices]

    def get_sequence(self, order: int) -> list[float]:
        """Get the Walsh-Hadamard sequence of given order (starts at 0)."""
        if order > self.max_order:
            self.set_max_order(order)

        return self.sequences[order]


def _get_transformation_matrix(n):
    """Get a 2^n x 2^n Walsh-Hadamard matrix with elements in [0, 1]."""

    from qiskit.circuit.library import HGate

    had = np.array(HGate()).real * np.sqrt(2)

    def recurse(matrix, m):
        # we build the matrix recursively, adding one Hadamard kronecker product per recursion
        if m == 1:
            # finally, map to [0, 1] by the mapping (1 - H) / 2
            return ((1 - matrix) / 2).astype(int)

        return recurse(np.kron(had, matrix), m - 1)

    return recurse(had, n)


def _bitflips_to_timings(row):
    num = len(row)
    distances = []
    count = 0
    last = 0  # start in no flip state
    for el in row:
        if el == last:
            count += 1
        else:
            distances.append(count / num)
            last = el
            count = 1

    distances.append(count / num)

    if len(distances) % 2 == 0:
        return distances + [0]

    return distances


def _gate_length_variance(target: Target) -> float:
    max_length, min_length = None, None

    for gate, properties in target.items():
        if gate not in ["cx", "cz", "ecr"]:
            continue

        for prop in properties.values():
            duration = prop.duration
            if max_length is None or max_length < duration:
                max_length = duration
            if min_length is None or min_length > duration:
                min_length = duration

    # it could be that there are no 2q gates available, in this
    # case we just return 0, which will mean we join all idle times
    if max_length is None or min_length is None:
        return 0

    return max_length - min_length



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Utility functions for generating random circuits."""
import numpy as np
from rustworkx import PyDiGraph

from qiskit.circuit import (
    ClassicalRegister,
    QuantumCircuit,
    CircuitInstruction,
)
from qiskit.circuit import Reset
from qiskit.circuit.library import standard_gates
from qiskit.circuit.exceptions import CircuitError
from qiskit.quantum_info.operators.symplectic.clifford_circuits import _BASIS_1Q, _BASIS_2Q

# (Gate class, number of qubits, number of parameters)
gates_1q_data = [
    (standard_gates.IGate, 1, 0),
    (standard_gates.SXGate, 1, 0),
    (standard_gates.XGate, 1, 0),
    (standard_gates.RZGate, 1, 1),
    (standard_gates.RGate, 1, 2),
    (standard_gates.HGate, 1, 0),
    (standard_gates.PhaseGate, 1, 1),
    (standard_gates.RXGate, 1, 1),
    (standard_gates.RYGate, 1, 1),
    (standard_gates.SGate, 1, 0),
    (standard_gates.SdgGate, 1, 0),
    (standard_gates.SXdgGate, 1, 0),
    (standard_gates.TGate, 1, 0),
    (standard_gates.TdgGate, 1, 0),
    (standard_gates.UGate, 1, 3),
    (standard_gates.U1Gate, 1, 1),
    (standard_gates.U2Gate, 1, 2),
    (standard_gates.U3Gate, 1, 3),
    (standard_gates.YGate, 1, 0),
    (standard_gates.ZGate, 1, 0),
]

gates_2q_data = [
    (standard_gates.CXGate, 2, 0),
    (standard_gates.DCXGate, 2, 0),
    (standard_gates.CHGate, 2, 0),
    (standard_gates.CPhaseGate, 2, 1),
    (standard_gates.CRXGate, 2, 1),
    (standard_gates.CRYGate, 2, 1),
    (standard_gates.CRZGate, 2, 1),
    (standard_gates.CSXGate, 2, 0),
    (standard_gates.CUGate, 2, 4),
    (standard_gates.CU1Gate, 2, 1),
    (standard_gates.CU3Gate, 2, 3),
    (standard_gates.CYGate, 2, 0),
    (standard_gates.CZGate, 2, 0),
    (standard_gates.RXXGate, 2, 1),
    (standard_gates.RYYGate, 2, 1),
    (standard_gates.RZZGate, 2, 1),
    (standard_gates.RZXGate, 2, 1),
    (standard_gates.XXMinusYYGate, 2, 2),
    (standard_gates.XXPlusYYGate, 2, 2),
    (standard_gates.ECRGate, 2, 0),
    (standard_gates.CSGate, 2, 0),
    (standard_gates.CSdgGate, 2, 0),
    (standard_gates.SwapGate, 2, 0),
    (standard_gates.iSwapGate, 2, 0),
]


def random_circuit_from_graph(
    interaction_graph,
    min_2q_gate_per_edge=1,
    max_operands=2,
    measure=False,
    conditional=False,
    reset=False,
    seed=None,
    insert_1q_oper=True,
    prob_conditional=0.1,
    prob_reset=0.1,
):
    """Generate random circuit of arbitrary size and form which strictly respects the interaction
    graph passed as argument. Interaction Graph is a graph G=(V, E) where V are the qubits in the
    circuit, and, E is the set of two-qubit gate interactions between two particular qubits in the
    circuit.

    This function will generate a random circuit by randomly selecting 1Q and 2Q gates from the set
    of standard gates in :mod:`qiskit.circuit.library.standard_gates`. The user can attach a numerical
    value as a metadata to the edge of the graph indicating the edge weight for that particular edge,
    These edge weights would be normalized to the probabilities of the edges getting selected.
    If all the edge weights are passed as `None`, then the probability of each qubit-pair of getting
    selected is set to 1/N, where `N` is the number of edges in the `interaction_graph` passed in,
    i.e each edge is drawn uniformly. If any weight of an edge is set as zero, that particular
    edge will not be included in the output circuit.

    Passing a list of tuples of control qubit, target qubit and associated probability is also
    acceptable as a way to specify an interaction graph.

    If numerical values are present as probabilities but some/any are None, or negative, when
    `max_operands=2` this will raise a ValueError.

    If `max_operands` is set to 1, then there are no 2Q operations, so no need to take care
    of the edges, in such case the function will return a circuit from the `random_circuit` function,
    which would be passed with the `max_operands` as 1.

    If `max_operands` is set to 2, then in every while-iteration 2Q gates are chosen randomly, and
    qubit-pairs which exist in the input interaction graph are also chosen randomly based on the
    probability attached to the qubit-pair. Then in a for-iteration 2Q gates are applied to the
    randomly chosen qubit-pairs with the aim to reach the count of 2Q on any qubit-pair to
    `min_2q_gate_per_edge` criteria as soon as possible within a particular iteration. Now, if
    some qubits are still idle after applying 2Q gates for that particular iteration, then randomly
    chosen 1Q gates are applied to those idle qubits if `insert_1q_oper` is set to True.

    Example:

    .. plot::
       :alt: Pass in interaction graph and minimum 2Q gate per edge as a bare minimum.
       :include-source:

       from qiskit.circuit.random.utils import random_circuit_from_graph
       import rustworkx as rx
       pydi_graph = rx.PyDiGraph()
       pydi_graph.add_nodes_from(range(7))
       cp_map = [(0, 1, 0.18), (1, 2, 0.15), (2, 3, 0.15), (3, 4, 0.22), (4, 5, 0.13), (5, 6, 0.17)]
       pydi_graph.add_edges_from(cp_map)
       # cp_map can be passed in directly as interaction_graph
       qc = random_circuit_from_graph(pydi_graph, min_2q_gate_per_edge=2, seed=12345)
       qc.draw(output='mpl')

    Args:
        interaction_graph (PyGraph | PyDiGraph | List[Tuple[int, int, float]]): Interaction Graph
        min_2q_gate_per_edge (int): Minimum number of times every qubit-pair must be used
            in the random circuit. (optional, default:1)
        max_operands (int): maximum qubit operands of each gate(should be 1 or 2)
            (optional, default:2)
        measure (bool): if True, measure all qubits at the end. (optional, default: False)
        conditional (bool): if True, insert middle measurements and conditionals.
            (optional, default: False)
        reset (bool): if True, insert middle resets. (optional, default: False)
        seed (int): sets random seed. (If `None`, a random seed is chosen) (optional)
        insert_1q_oper (bool): Insert 1Q operations to the circuit. (optional, default: True)
        prob_conditional (float): Probability less than 1.0, this is used to control the occurrence
            of conditionals in the circuit. (optional, default: 0.1)
        prob_reset (float): Probability less than 1.0, this is used to control the occurrence of
            reset in the circuit. (optional, default: 0.1)

    Returns:
        QuantumCircuit: constructed circuit

    Raises:
        CircuitError: When `max_operands` is not 1 or 2.
        CircuitError: When `max_operands` is set to 1, but no 1Q operations are allowed by setting
            `insert_1q_oper` to false.
        CircuitError: When the interaction graph has no edges, so only 1Q gates are possible in
            the circuit, but `insert_1q_oper` is set to False.
        CircuitError: When an invalid interaction graph object is passed.
        ValueError: Given `max_operands=2`, when any edge have probability `None` but not all, or any
            of the probabilities are negative.
    """

    # max_operands should be 1 or 2
    if max_operands not in {1, 2}:
        raise CircuitError("`max_operands` should be either 1 or 2")

    if max_operands == 1 and not insert_1q_oper:
        raise CircuitError(
            "`max_operands` of 1 means only 1Q gates are allowed, but `insert_1q_oper` is False"
        )

    # Declaring variables so lint doesn't complaint.
    num_qubits = 0
    num_edges = None
    edge_list = None
    edges_probs = None

    if isinstance(interaction_graph, list):
        num_edges = len(interaction_graph)
        edge_list = []
        edges_probs = []
        for ctrl, trgt, prob in interaction_graph:
            edge_list.append((ctrl, trgt))
            edges_probs.append(prob)

            if ctrl > num_qubits:
                num_qubits = ctrl

            if trgt > num_qubits:
                num_qubits = trgt

        num_qubits += 1  # ctrl, trgt are qubit indices.
        edge_list = np.array(edge_list)
        edges_probs = np.array(edges_probs)
    elif isinstance(interaction_graph, PyDiGraph):
        num_qubits = interaction_graph.num_nodes()
        num_edges = interaction_graph.num_edges()
        edge_list = np.array(interaction_graph.edge_list())
        edges_probs = np.array(interaction_graph.edges())
    else:
        raise CircuitError("Invalid interaction graph object has been passed")

    if num_qubits == 0:
        return QuantumCircuit()

    max_operands = max_operands if num_qubits > max_operands else num_qubits

    if num_edges == 0 and not insert_1q_oper:
        raise CircuitError(
            "There are no edges in the `interaction_graph` so, there could be only 1Q gates, "
            "however `insert_1q_oper` is set to `False`"
        )

    if num_edges == 0 or max_operands == 1:
        # If there is no edge then there could be no 2Q operation
        # or, if only 1Q operations are allowed, then there is no
        # point in considering edges.
        return random_circuit(
            num_qubits=num_qubits,
            depth=min_2q_gate_per_edge,
            max_operands=1,
            measure=measure,
            conditional=conditional,
            reset=reset,
            seed=seed,
        )

    # If any edge weight is zero, just remove that edge from the edge_list
    if 0 in edges_probs:
        _mask = edges_probs != 0
        edges_probs = edges_probs[_mask]
        edge_list = edge_list[_mask]

        # Update 'num_edges'
        num_edges = len(edge_list)

    # Now, zeros are filtered out in above if-block.
    # Now, If none of the edges_probs are `None`
    if all(edges_probs):

        # edge weights in interaction_graph must be positive
        for prob in edges_probs:
            if prob < 0:
                raise ValueError("Probabilities cannot be negative")

        # Normalize edge weights if not already normalized.
        probs_sum = edges_probs.sum()
        if not np.isclose(probs_sum, 1.000, rtol=0.001):
            edges_probs = edges_probs / probs_sum
    # If any of the values of the probability is None, then it would raise an error.
    elif any(edges_probs):
        raise ValueError(
            "Some of the probabilities of a qubit-pair getting selected is `None`"
            " It should either be all `None` or all positive numerical weights. "
        )

    # If all edge weights are none, assume the weight of each edge to be 1/N.
    elif None in edges_probs:
        edges_probs = np.ones(num_edges) / num_edges

    gates_2q = np.array(
        gates_2q_data,
        dtype=[("class", object), ("num_qubits", np.int64), ("num_params", np.int64)],
    )

    if insert_1q_oper:
        gates_1q = np.array(gates_1q_data, dtype=gates_2q.dtype)

    qc = QuantumCircuit(num_qubits)

    if measure or conditional:
        cr = ClassicalRegister(num_qubits, "c")
        qc.add_register(cr)

    if seed is None:
        seed = np.random.randint(0, np.iinfo(np.int32).max)
    rng = np.random.default_rng(seed)

    qubits = np.array(qc.qubits, dtype=object, copy=True)

    edges_used = {tuple(edge): 0 for edge in edge_list}

    # Declaring variables, so that lint doesn't complaint.
    reset_2q = None
    cond_val_2q = None
    cond_val_1q = None

    # If conditional is not required, there is no need to calculate random numbers.
    # But, since the variables are required in the for-loop so let's get an array of false.
    if not conditional:
        cond_1q = np.zeros(num_qubits, dtype=bool)
        cond_2q = np.zeros(num_edges, dtype=bool)

    # Similarly, if resets are not required, then since, the variable is required
    # in the for-loop, let's get an array of false.
    if not reset:
        reset_2q = np.zeros(num_edges, dtype=bool)

    # This loop will keep on applying gates to qubits until every qubit-pair
    # has 2Q operations applied at-least `min_2q_gate_per_edge` times.
    while edges_used:

        # For any given while-iteration, this is a set of qubits not having any 2Q gates.
        # Qubits in this set will have 1Q gates, if `insert_1q_oper` is True.
        qubit_idx_idle = set(range(num_qubits))

        # normalized edge weights represent the probability with which each qubit-pair
        # is inserted into the circuit.
        edge_choices = rng.choice(
            edge_list,
            size=num_edges,
            replace=True,
            p=edges_probs,
        )
        gate_choices = rng.choice(gates_2q, size=num_edges, replace=True)

        cumsum_params = np.cumsum(gate_choices["num_params"], dtype=np.int64)
        parameters = rng.uniform(0, 2 * np.pi, size=cumsum_params[-1])

        # If reset is required, then, generating a random boolean matrix of
        # num_edges x 2, corresponding to probable reset on both control and
        # target qubits of the edge from the edge_list.
        if reset:
            reset_2q = rng.random(size=(num_edges, 2)) < prob_reset

        if conditional:
            cond_2q = rng.random(size=len(gate_choices)) < prob_conditional
            cond_val_2q = rng.integers(0, 1 << min(num_qubits, 63), size=np.count_nonzero(cond_2q))
            clbit_2q_idx = 0

        for gate, num_gate_params, edge, is_cond_2q, is_rst in zip(
            gate_choices["class"],
            gate_choices["num_params"],
            edge_choices,
            cond_2q,
            reset_2q,
        ):

            control_qubit, target_qubit = tuple(edge)

            # For every edge there are two probabilistically generated boolean values corresponding
            # to control, target qubits of the edge
            # an idle qubit for a particular iteration on which reset is applied is considered idle.

            if reset:
                is_rst_control, is_rst_target = is_rst
                rst_oper = Reset()
                if is_rst_control:
                    qc._append(
                        CircuitInstruction(
                            operation=rst_oper,
                            qubits=[qubits[control_qubit]],
                        )
                    )

                if is_rst_target:
                    qc._append(
                        CircuitInstruction(
                            operation=rst_oper,
                            qubits=[qubits[target_qubit]],
                        )
                    )

            params = parameters[:num_gate_params]
            parameters = parameters[num_gate_params:]
            current_instr = gate(*params)

            if is_cond_2q:
                qc.measure(qc.qubits, cr)
                # The condition values are required to be bigints, not Numpy's fixed-width type.
                with qc.if_test((cr, int(cond_val_2q[clbit_2q_idx]))):
                    clbit_2q_idx += 1
                    qc.append(
                        CircuitInstruction(
                            operation=current_instr,
                            qubits=[qubits[control_qubit], qubits[target_qubit]],
                        )
                    )
            else:
                qc._append(
                    CircuitInstruction(
                        operation=current_instr,
                        qubits=[qubits[control_qubit], qubits[target_qubit]],
                    )
                )

            # Removing the qubits that have been applied with 2Q gates from the
            # set of idle qubits for that while-iteration.
            qubit_idx_idle = qubit_idx_idle - set(edge)

            # Update the number of occurrences of the edge in the circuit.
            _temp_edge = (control_qubit, target_qubit)
            if _temp_edge in edges_used:
                edges_used[_temp_edge] += 1
                if edges_used[_temp_edge] >= min_2q_gate_per_edge:
                    del edges_used[_temp_edge]

        if insert_1q_oper:
            num_unused_qubits = len(qubit_idx_idle)
            if not num_unused_qubits == 0:

                # Calculating for conditionals to make even the 1Q operations
                # probabilistically conditional.
                if conditional:
                    cond_1q = rng.random(size=num_unused_qubits) < prob_conditional
                    cond_val_1q = rng.integers(
                        0, 1 << min(num_qubits, 63), size=np.count_nonzero(cond_1q)
                    )
                    clbit_1q_idx = 0

                # Some extra 1Q Gate in to fill qubits which are still idle for this
                # particular while iteration.
                extra_1q_gates = rng.choice(gates_1q, size=num_unused_qubits, replace=True)

                cumsum_params = np.cumsum(extra_1q_gates["num_params"], dtype=np.int64)
                parameters_1q = rng.uniform(0, 2 * np.pi, size=cumsum_params[-1])

                for gate_1q, num_gate_params, qubit_idx, is_cond_1q in zip(
                    extra_1q_gates["class"],
                    extra_1q_gates["num_params"],
                    qubit_idx_idle,
                    cond_1q,
                ):
                    params_1q = parameters_1q[:num_gate_params]
                    parameters_1q = parameters_1q[num_gate_params:]
                    current_instr = gate_1q(*params_1q)

                    if is_cond_1q:
                        qc.measure(qc.qubits, cr)
                        # The condition values are required to be bigints, not Numpy's fixed-width type.
                        with qc.if_test((cr, int(cond_val_1q[clbit_1q_idx]))):
                            clbit_1q_idx += 1
                            qc.append(
                                CircuitInstruction(
                                    operation=current_instr,
                                    qubits=[qubits[qubit_idx]],
                                )
                            )
                    else:
                        qc._append(
                            CircuitInstruction(
                                operation=current_instr,
                                qubits=[qubits[qubit_idx]],
                            )
                        )

    if measure:
        qc.measure(qc.qubits, cr)

    return qc


def random_circuit(
    num_qubits,
    depth,
    max_operands=4,
    measure=False,
    conditional=False,
    reset=False,
    seed=None,
    num_operand_distribution: dict = None,
):
    """Generate random circuit of arbitrary size and form.

    This function will generate a random circuit by randomly selecting gates
    from the set of standard gates in :mod:`qiskit.circuit.library.standard_gates`. For example:

    .. plot::
       :alt: Circuit diagram output by the previous code.
       :include-source:

       from qiskit.circuit.random import random_circuit

       circ = random_circuit(2, 2, measure=True)
       circ.draw(output='mpl')

    Args:
        num_qubits (int): number of quantum wires
        depth (int): layers of operations (i.e. critical path length)
        max_operands (int): maximum qubit operands of each gate (between 1 and 4)
        measure (bool): if True, measure all qubits at the end
        conditional (bool): if True, insert middle measurements and conditionals
        reset (bool): if True, insert middle resets
        seed (int): sets random seed (optional)
        num_operand_distribution (dict): a distribution of gates that specifies the ratio
            of 1-qubit, 2-qubit, 3-qubit, ..., n-qubit gates in the random circuit. Expect a
            deviation from the specified ratios that depends on the size of the requested
            random circuit. (optional)

    Returns:
        QuantumCircuit: constructed circuit

    Raises:
        CircuitError: when invalid options given
    """
    if seed is None:
        seed = np.random.randint(0, np.iinfo(np.int32).max)
    rng = np.random.default_rng(seed)

    if num_operand_distribution:
        if min(num_operand_distribution.keys()) < 1 or max(num_operand_distribution.keys()) > 4:
            raise CircuitError("'num_operand_distribution' must have keys between 1 and 4")
        for key, prob in num_operand_distribution.items():
            if key > num_qubits and prob != 0.0:
                raise CircuitError(
                    f"'num_operand_distribution' cannot have {key}-qubit gates"
                    f" for circuit with {num_qubits} qubits"
                )
        num_operand_distribution = dict(sorted(num_operand_distribution.items()))

    if not num_operand_distribution and max_operands:
        if max_operands < 1 or max_operands > 4:
            raise CircuitError("max_operands must be between 1 and 4")
        max_operands = max_operands if num_qubits > max_operands else num_qubits
        rand_dist = rng.dirichlet(
            np.ones(max_operands)
        )  # This will create a random distribution that sums to 1
        num_operand_distribution = {i + 1: rand_dist[i] for i in range(max_operands)}
        num_operand_distribution = dict(sorted(num_operand_distribution.items()))

    # Here we will use np.isclose() because very rarely there might be floating
    # point precision errors
    if not np.isclose(sum(num_operand_distribution.values()), 1):
        raise CircuitError("The sum of all the values in 'num_operand_distribution' is not 1.")

    if num_qubits == 0:
        return QuantumCircuit()

    gates_3q = [
        (standard_gates.CCXGate, 3, 0),
        (standard_gates.CSwapGate, 3, 0),
        (standard_gates.CCZGate, 3, 0),
        (standard_gates.RCCXGate, 3, 0),
    ]

    gates_4q = [
        (standard_gates.C3SXGate, 4, 0),
        (standard_gates.RC3XGate, 4, 0),
    ]

    gates_1q = np.array(
        gates_1q_data + [(Reset, 1, 0)] if reset else gates_1q_data,
        dtype=[("class", object), ("num_qubits", np.int64), ("num_params", np.int64)],
    )

    gates_2q = np.array(gates_2q_data, dtype=gates_1q.dtype)
    gates_3q = np.array(gates_3q, dtype=gates_1q.dtype)
    gates_4q = np.array(gates_4q, dtype=gates_1q.dtype)

    all_gate_lists = [gates_1q, gates_2q, gates_3q, gates_4q]

    # Here we will create a list 'gates_to_consider' that will have a
    # subset of different n-qubit gates and will also create a list for
    # ratio (or probability) for each gates
    gates_to_consider = []
    distribution = []
    for n_qubits, ratio in num_operand_distribution.items():
        gate_list = all_gate_lists[n_qubits - 1]
        gates_to_consider.extend(gate_list)
        distribution.extend([ratio / len(gate_list)] * len(gate_list))

    gates = np.array(gates_to_consider, dtype=gates_1q.dtype)

    qc = QuantumCircuit(num_qubits)

    if measure or conditional:
        cr = ClassicalRegister(num_qubits, "c")
        qc.add_register(cr)

    qubits = np.array(qc.qubits, dtype=object, copy=True)

    # Counter to keep track of number of different gate types
    counter = np.zeros(len(all_gate_lists) + 1, dtype=np.int64)
    total_gates = 0

    # Apply arbitrary random operations in layers across all qubits.
    for layer_number in range(depth):
        # We generate all the randomness for the layer in one go, to avoid many separate calls to
        # the randomization routines, which can be fairly slow.

        # This reliably draws too much randomness, but it's less expensive than looping over more
        # calls to the rng. After, trim it down by finding the point when we've used all the qubits.

        # Due to the stochastic nature of generating a random circuit, the resulting ratios
        # may not precisely match the specified values from `num_operand_distribution`. Expect
        # greater deviations from the target ratios in quantum circuits with fewer qubits and
        # shallower depths, and smaller deviations in larger and deeper quantum circuits.
        # For more information on how the distribution changes with number of qubits and depth
        # refer to the pull request #12483 on Qiskit GitHub.

        gate_specs = rng.choice(gates, size=len(qubits), p=distribution)
        cumulative_qubits = np.cumsum(gate_specs["num_qubits"], dtype=np.int64)

        # Efficiently find the point in the list where the total gates would use as many as
        # possible of, but not more than, the number of qubits in the layer.  If there's slack, fill
        # it with 1q gates.
        max_index = np.searchsorted(cumulative_qubits, num_qubits, side="right")
        gate_specs = gate_specs[:max_index]

        slack = num_qubits - cumulative_qubits[max_index - 1]

        # Updating the counter for 1-qubit, 2-qubit, 3-qubit and 4-qubit gates
        gate_qubits = gate_specs["num_qubits"]
        counter += np.bincount(gate_qubits, minlength=len(all_gate_lists) + 1)

        total_gates += len(gate_specs)

        # Slack handling loop, this loop will add gates to fill
        # the slack while respecting the 'num_operand_distribution'
        while slack > 0:
            gate_added_flag = False

            for key, dist in sorted(num_operand_distribution.items(), reverse=True):
                if slack >= key and counter[key] / total_gates < dist:
                    gate_to_add = np.array(
                        all_gate_lists[key - 1][rng.integers(0, len(all_gate_lists[key - 1]))]
                    )
                    gate_specs = np.hstack((gate_specs, gate_to_add))
                    counter[key] += 1
                    total_gates += 1
                    slack -= key
                    gate_added_flag = True
            if not gate_added_flag:
                break

        # For efficiency in the Python loop, this uses Numpy vectorization to pre-calculate the
        # indices into the lists of qubits and parameters for every gate, and then suitably
        # randomizes those lists.
        q_indices = np.empty(len(gate_specs) + 1, dtype=np.int64)
        p_indices = np.empty(len(gate_specs) + 1, dtype=np.int64)
        q_indices[0] = p_indices[0] = 0
        np.cumsum(gate_specs["num_qubits"], out=q_indices[1:])
        np.cumsum(gate_specs["num_params"], out=p_indices[1:])
        parameters = rng.uniform(0, 2 * np.pi, size=p_indices[-1])
        rng.shuffle(qubits)

        # We've now generated everything we're going to need.  Now just to add everything.  The
        # conditional check is outside the two loops to make the more common case of no conditionals
        # faster, since in Python we don't have a compiler to do this for us.
        if conditional and layer_number != 0:
            is_conditional = rng.random(size=len(gate_specs)) < 0.1
            condition_values = rng.integers(
                0, 1 << min(num_qubits, 63), size=np.count_nonzero(is_conditional)
            )
            c_ptr = 0
            for gate, q_start, q_end, p_start, p_end, is_cond in zip(
                gate_specs["class"],
                q_indices[:-1],
                q_indices[1:],
                p_indices[:-1],
                p_indices[1:],
                is_conditional,
            ):
                operation = gate(*parameters[p_start:p_end])

                if is_cond:
                    qc.measure(qc.qubits, cr)
                    # The condition values are required to be bigints, not Numpy's fixed-width type.
                    with qc.if_test((cr, int(condition_values[c_ptr]))):
                        c_ptr += 1
                        qc.append(
                            CircuitInstruction(operation=operation, qubits=qubits[q_start:q_end])
                        )
                else:
                    qc._append(
                        CircuitInstruction(operation=operation, qubits=qubits[q_start:q_end])
                    )
        else:
            for gate, q_start, q_end, p_start, p_end in zip(
                gate_specs["class"], q_indices[:-1], q_indices[1:], p_indices[:-1], p_indices[1:]
            ):
                operation = gate(*parameters[p_start:p_end])
                qc._append(CircuitInstruction(operation=operation, qubits=qubits[q_start:q_end]))
    if measure:
        qc.measure(qc.qubits, cr)

    return qc


def random_clifford_circuit(num_qubits, num_gates, gates="all", seed=None):
    """Generate a pseudo-random Clifford circuit.

    This function will generate a Clifford circuit by randomly selecting the chosen amount of Clifford
    gates from the set of standard gates in :mod:`qiskit.circuit.library.standard_gates`. For example:

    .. plot::
       :alt: Circuit diagram output by the previous code.
       :include-source:

       from qiskit.circuit.random import random_clifford_circuit

       circ = random_clifford_circuit(num_qubits=2, num_gates=6)
       circ.draw(output='mpl')

    Args:
        num_qubits (int): number of quantum wires.
        num_gates (int): number of gates in the circuit.
        gates (list[str]): optional list of Clifford gate names to randomly sample from.
            If ``"all"`` (default), use all Clifford gates in the standard library.
        seed (int | np.random.Generator): sets random seed/generator (optional).

    Returns:
        QuantumCircuit: constructed circuit
    """

    gates_1q = list(set(_BASIS_1Q.keys()) - {"v", "w", "id", "iden", "sinv"})
    gates_2q = list(_BASIS_2Q.keys())

    if gates == "all":
        if num_qubits == 1:
            gates = gates_1q
        else:
            gates = gates_1q + gates_2q

    instructions = {
        "i": (standard_gates.IGate(), 1),
        "x": (standard_gates.XGate(), 1),
        "y": (standard_gates.YGate(), 1),
        "z": (standard_gates.ZGate(), 1),
        "h": (standard_gates.HGate(), 1),
        "s": (standard_gates.SGate(), 1),
        "sdg": (standard_gates.SdgGate(), 1),
        "sx": (standard_gates.SXGate(), 1),
        "sxdg": (standard_gates.SXdgGate(), 1),
        "cx": (standard_gates.CXGate(), 2),
        "cy": (standard_gates.CYGate(), 2),
        "cz": (standard_gates.CZGate(), 2),
        "swap": (standard_gates.SwapGate(), 2),
        "iswap": (standard_gates.iSwapGate(), 2),
        "ecr": (standard_gates.ECRGate(), 2),
        "dcx": (standard_gates.DCXGate(), 2),
    }

    if isinstance(seed, np.random.Generator):
        rng = seed
    else:
        rng = np.random.default_rng(seed)

    samples = rng.choice(gates, num_gates)

    circ = QuantumCircuit(num_qubits)

    for name in samples:
        gate, nqargs = instructions[name]
        qargs = rng.choice(range(num_qubits), nqargs, replace=False).tolist()
        circ.append(gate, qargs, copy=False)

    return circ



# This code is part of Qiskit.
#
# (C) Copyright IBM 2021.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Two-qubit XX+YY gate."""

from __future__ import annotations

import math
from cmath import exp
from typing import Optional

import numpy

from qiskit.circuit.gate import Gate
from qiskit.circuit.parameterexpression import ParameterValueType, ParameterExpression
from qiskit._accelerate.circuit import StandardGate


class XXPlusYYGate(Gate):
    r"""XX+YY interaction gate.

    A 2-qubit parameterized XX+YY interaction, also known as an XY gate. Its action is to induce
    a coherent rotation by some angle between :math:`|01\rangle` and :math:`|10\rangle`.

    **Circuit Symbol:**

    .. code-block:: text

             ┌───────────────┐
        q_0: ┤0              ├
             │  (XX+YY)(θ,β) │
        q_1: ┤1              ├
             └───────────────┘

    **Matrix Representation:**

    .. math::

        \newcommand{\rotationangle}{\frac{\theta}{2}}

        R_{XX+YY}(\theta, \beta)\ q_0, q_1 =
          RZ_0(-\beta) \cdot \exp\left(-i \frac{\theta}{2} \frac{XX+YY}{2}\right) \cdot RZ_0(\beta) =
            \begin{pmatrix}
                1 & 0 & 0 & 0  \\
                0 & \cos\left(\rotationangle\right) & -i\sin\left(\rotationangle\right)e^{-i\beta} & 0 \\
                0 & -i\sin\left(\rotationangle\right)e^{i\beta} & \cos\left(\rotationangle\right) & 0 \\
                0 & 0 & 0 & 1
            \end{pmatrix}

    .. note::

        In Qiskit's convention, higher qubit indices are more significant
        (little endian convention). In the above example we apply the gate
        on (q_0, q_1) which results in adding the (optional) phase defined
        by :math:`\beta` on q_0. Instead, if we apply it on (q_1, q_0), the
        phase is added on q_1. If :math:`\beta` is set to its default value
        of :math:`0`, the gate is equivalent in big and little endian.

        .. code-block:: text

                 ┌───────────────┐
            q_0: ┤1              ├
                 │  (XX+YY)(θ,β) │
            q_1: ┤0              ├
                 └───────────────┘

        .. math::

            \newcommand{\rotationangle}{\frac{\theta}{2}}

            R_{XX+YY}(\theta, \beta)\ q_0, q_1 =
            RZ_1(-\beta) \cdot \exp\left(-i \frac{\theta}{2} \frac{XX+YY}{2}\right) \cdot RZ_1(\beta) =
                \begin{pmatrix}
                    1 & 0 & 0 & 0  \\
                    0 & \cos\left(\rotationangle\right) &
                    -i\sin\left(\rotationangle\right)e^{i\beta} & 0 \\
                    0 & -i\sin\left(\rotationangle\right)e^{-i\beta} &
                    \cos\left(\rotationangle\right) & 0 \\
                    0 & 0 & 0 & 1
                \end{pmatrix}
    """

    _standard_gate = StandardGate.XXPlusYY

    def __init__(
        self,
        theta: ParameterValueType,
        beta: ParameterValueType = 0,
        label: Optional[str] = "(XX+YY)",
    ):
        """Create new XX+YY gate.

        Args:
            theta: The rotation angle.
            beta: The phase angle.
            label: The label of the gate.
        """
        super().__init__("xx_plus_yy", 2, [theta, beta], label=label)

    def _define(self):
        """Default definition"""
        # pylint: disable=cyclic-import
        from qiskit.circuit import QuantumCircuit

        #      ┌───────┐┌───┐      ┌───┐┌──────────┐┌───┐┌─────┐┌────────┐
        # q_0: ┤ Rz(β) ├┤ S ├──────┤ X ├┤ Ry(-θ/2) ├┤ X ├┤ Sdg ├┤ Rz(-β) ├─────
        #      └┬─────┬┘├───┴┐┌───┐└─┬─┘├──────────┤└─┬─┘├─────┤└┬──────┬┘┌───┐
        # q_1: ─┤ Sdg ├─┤ √X ├┤ S ├──■──┤ Ry(-θ/2) ├──■──┤ Sdg ├─┤ √Xdg ├─┤ S ├
        #       └─────┘ └────┘└───┘     └──────────┘     └─────┘ └──────┘ └───┘

        self.definition = QuantumCircuit._from_circuit_data(
            StandardGate.XXPlusYY._get_definition(self.params), add_regs=True, name=self.name
        )

    def control(
        self,
        num_ctrl_qubits: int = 1,
        label: str | None = None,
        ctrl_state: str | int | None = None,
        annotated: bool | None = None,
    ):
        """Return a (multi-)controlled-(XX+YY) gate.

        Args:
            num_ctrl_qubits: number of control qubits.
            label: An optional label for the gate [Default: ``None``]
            ctrl_state: control state expressed as integer,
                string (e.g.``'110'``), or ``None``. If ``None``, use all 1s.
            annotated: indicates whether the controlled gate should be implemented
                as an annotated gate. If ``None``, this is set to ``True`` if
                the gate contains free parameters, in which case it cannot
                yet be synthesized.

        Returns:
            ControlledGate: controlled version of this gate.
        """
        if annotated is None:
            annotated = any(isinstance(p, ParameterExpression) for p in self.params)

        gate = super().control(
            num_ctrl_qubits=num_ctrl_qubits,
            label=label,
            ctrl_state=ctrl_state,
            annotated=annotated,
        )
        return gate

    def inverse(self, annotated: bool = False):
        """Return inverse XX+YY gate (i.e. with the negative rotation angle and same phase angle).

        Args:
            annotated: when set to ``True``, this is typically used to return an
                :class:`.AnnotatedOperation` with an inverse modifier set instead of a concrete
                :class:`.Gate`. However, for this class this argument is ignored as the inverse
                of this gate is always a :class:`.XXPlusYYGate` with inverse
                parameter values.

        Returns:
            XXPlusYYGate: inverse gate.
        """
        return XXPlusYYGate(-self.params[0], self.params[1])

    def __array__(self, dtype=None, copy=None):
        """Return a numpy.array for the XX+YY gate."""
        if copy is False:
            raise ValueError("unable to avoid copy while creating an array as requested")
        half_theta = float(self.params[0]) / 2
        beta = float(self.params[1])
        cos = math.cos(half_theta)
        sin = math.sin(half_theta)
        return numpy.array(
            [
                [1, 0, 0, 0],
                [0, cos, -1j * sin * exp(-1j * beta), 0],
                [0, -1j * sin * exp(1j * beta), cos, 0],
                [0, 0, 0, 1],
            ],
            dtype=dtype,
        )

    def power(self, exponent: float, annotated: bool = False):
        theta, beta = self.params
        return XXPlusYYGate(exponent * theta, beta)

    def __eq__(self, other):
        if isinstance(other, XXPlusYYGate):
            return self._compare_parameters(other)
        return False
